// graph-tool -- a general graph modification and manipulation thingy
//
// Copyright (C) 2006-2026 Tiago de Paula Peixoto <tiago@skewed.de>
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef OPENMP_HH
#define OPENMP_HH
#include "config.h"

#include <cstddef>

#ifdef _OPENMP
# include <omp.h>
#endif

#pragma GCC visibility push(default)

#ifndef _OPENMP
constexpr
#endif
inline int get_num_threads()
{
#ifdef _OPENMP
    return omp_get_max_threads();
#else
    return 1;
#endif
}

#ifndef _OPENMP
constexpr
#endif
inline int get_thread_num()
{
#ifdef _OPENMP
    return omp_get_thread_num();
#else
    return 0;
#endif
}

size_t get_openmp_min_thresh();
void set_openmp_min_thresh(size_t);

#pragma GCC visibility pop

#endif // OPENMP_HH
